%% Solutions for BGGN 260 Computational Lab and Homework 4, Fall 2019
close all; clear all

%% Hodgkin-Huxley Model

% Constants
C_m  =   1.0; % membrane capacitance, in uF/cm^2
g_Na = 120.0; % maximum conducances, in mS/cm^2
g_K  =  36.0;
g_L  =   0.3;
E_Na =  45.0; % Nernst reversal potentials, in mV
E_K  = -82.0;
E_L  = -59.387;
E_Cl = -80.0; % inhibitory synapse
E_ex = -38.0; % excitatory synapse

% Channel gating kinetics
% Functions of membrane voltage
alpha_m = @(V) 0.1.*(V+45.0)./(1.0 - exp(-(V+45.0) ./ 10.0));
beta_m  = @(V) 4.0.*exp(-(V+70.0) ./ 18.0);
alpha_h = @(V) 0.07.*exp(-(V+70.0) ./ 20.0);
beta_h  = @(V) 1.0./(1.0 + exp(-(V+40.0) ./ 10.0));
alpha_n = @(V) 0.01.*(V+60.0)./(1.0 - exp(-(V+60.0) ./ 10.0));
beta_n  = @(V) 0.125.*exp(-(V+70) ./ 80.0);

% Alpha / Beta constants for inhibitory / excitatory synapses
alpha_r_i = 5.0;  % 1/(mM*ms)
beta_r_i  = 0.18; % 1/ms
alpha_r_e = 2.4;  % 1/(mM*ms)
beta_r_e  = 0.56; % 1/ms

% [T] equation for synapses
T_max_i = 1.5; % mM (inhibitory)
T_max_e = 1.0; % mM (excitatory)
K_p = 5.0; % mV
V_p = 7.0; % mV
T_i = @(V_pre) T_max_i ./ (1.0 + exp(-(V_pre - V_p)./K_p));
T_e = @(V_pre) T_max_e ./ (1.0 + exp(-(V_pre - V_p)./K_p));

% Differential gating equations
dm = @(V,m) alpha_m(V).*(1.0-m)-beta_m(V).*m;
dh = @(V,h) alpha_h(V).*(1.0-h)-beta_h(V).*h;
dn = @(V,n) alpha_n(V).*(1.0-n)-beta_n(V).*n;
dr_i = @(V,r) alpha_r_i.*T_i(V).*(1.0-r)-beta_r_i.*r; % V=V_pre, r is for pre
dr_e = @(V,r) alpha_r_e.*T_e(V).*(1.0-r)-beta_r_e.*r; % V=V_pre, r is for pre

% Membrane currents (in uA/cm^2)
I_Na = @(V,m,h) g_Na .* m.^3 .* h .* (V - E_Na);
I_K  = @(V, n)  g_K  .* n.^4      .* (V - E_K);
I_L  = @(V)     g_L               .* (V - E_L);
I_syn_i = @(V,g_GABA,r) r*g_GABA.*(V-E_Cl); % V=V_post, r is for pre
I_syn_e = @(V,g_Glu,r)  r*g_Glu .*(V-E_ex); % V=V_post, r is for pre

% Membrane voltage differential
dV = @(V, m, h, n, r_i, r_e, I_ext, g_GABA, g_Glu) ...
    (-I_Na(V,m,h)-I_K(V,n)-I_L(V)-I_syn_i(V,g_GABA,r_i)-I_syn_e(V,g_Glu,r_e)+I_ext) ./ C_m;


%% Now setup the fancy differential equations
% Run the differential equations for a single neuron
%
% This function is vectorizable. Every column in x is a cell. 
% The rows in x are V, m, h, n, r_i, and r_e.
%
% I_exts is a function handle that accepts the time and returns a row
% vector of current applied to each cell.
%
% g_GABA is a matrix of inhibitory synapses where each row is the
% presynaptic cell and the column is the postsynaptic cell.
% Same for g_Glu except that it is excitatory.
%
% This function works for a single cell when using ode45, but for more than
% one cell you need to use the 'd' function below.
d_single = @(t, x, I_ext, g_GABA, g_Glu) [
    dV(x(1,:),x(2,:),x(3,:),x(4,:),x(5,:),x(6,:),I_ext(t),g_GABA,g_Glu);
    dm(x(1,:),x(2,:));
    dh(x(1,:),x(3,:));
    dn(x(1,:),x(4,:));
    dr_i(x(1,:),x(5,:));
    dr_e(x(1,:),x(6,:))
    ];

% Run the differential equations for many neurons
%
% See above for definitions of I_exts, g_GABA, and g_Glu
%
% Just uses the single version, which is vectorizable, however since ode45
% needs column vectors this function reshapes the vector to work properly.
d = @(t, x, I_exts, g_GABA, g_Glu) ...
    reshape(d_single(t, reshape(x, 6, length(x)/6), I_exts, g_GABA, g_Glu), length(x), 1);

% Timespan
T = linspace(0, 500, 5000); % each division is 0.1 ms

% Function for network simulation
% [t, X] = network(...)
% returns time vector and X matrix
% each row of X is a timestep
% each column of X corresponds to V, m, h, n r_i, and r_e of a specific neuron. 
% column 1 = V of neuron 1, column 7 = V of neuron 2, etc.
network = @(I_exts, g_GABA, g_Glu) ...
    ode45(d, T, zeros(1, size(g_GABA, 1).*6), [], I_exts, g_GABA, g_Glu);   % Sets initial coditions for V, m, h, n,re_i, and re_e = 0



%% Example Two Neuron Network

% One neuron gets 15 uA (starting at 0ms) and the other gets 25 uA
I_exts_ = [15 25]; % uA/cm^2
I_exts = @(t) I_exts_;

% g_GABA and g_Glu matrices are all 0 for no connections
g_GABA = [0 0; 0 0];
g_Glu = [0 0; 0 0];

[t,X] = network(I_exts, g_GABA, g_Glu);

V = X(:,1:6:end);
%m   = X(:,2:6:end);
%h   = X(:,3:6:end);
%n   = X(:,4:6:end);
%r_i = X(:,5:6:end);
%r_e = X(:,6:6:end);

figure()
plot(t(4000:end), V(4000:end,:)) % plot last 100ms 
axis([400 500 min(min(V(4000:end,:)))-10 max(max(V(4000:end)))+10])
title('Example Network');
xlabel('Time (ms)');
ylabel('V_m (mV)');
legend('I_{ext} = 15\muA/cm^2', 'I_{ext} = 25\muA/cm^2', 'Location', 'SouthEast');
